// "The two-loop helicity amplitudes for qqb' -> V1V2 -> 4 leptons"
// T. Gehrmann, A. von Manteuffel, L. Tancredi
//
// qqvvamp package


#include "qqvvamp.h"
#include <exception>
#include <sstream>
#include <ginac/ginac.h>
#include <cln/float.h>
#include <cln/rational_ring.h>

using namespace GiNaC;
using namespace std;

// general conversion functions

inline GiNaC::numeric torealnumeric(const GiNaC::ex& x) {
  const numeric delta_im(1e-5); // ignore unexpected imaginary parts smaller than this
  ex n = x.evalf();
  if (!GiNaC::is_a<numeric>(n))
      throw runtime_error("failed numerical evaluation (polylog ?)");
  if (abs(imag_part(x)) > delta_im)
    cerr << "qqvvamp: unexpected imaginary part (polylog ?)" << endl;
  return ex_to<numeric>(real_part(n));
}

// instantiation specific functions

#ifdef ARBITRARY_PRECISION

GiNaC::numeric tonumeric(const cln::cl_F& x) {
  return numeric(cln::cl_N(x));
}

GiNaC::numeric tonumeric(const cln::cl_N& x) {
  return numeric(x);
}

double todouble(const cln::cl_F& x) {
  return double_approx(x);
}

inline cln::cl_F toclf(const GiNaC::ex& x) {
  cln::cl_N n = torealnumeric(x).to_cl_N();
  if (cln::instanceof(n, cln::cl_RA_ring))
    return cln::cl_float(As(cln::cl_RA)(n), cln::default_float_format);
  //n = cln::cl_float(1.0, cln::default_float_format) * n;
  return cln::cl_float(As(cln::cl_F)(n), cln::default_float_format);
}

template<> cln::cl_F qqvvamp<cln::cl_F>::n(const GiNaC::ex& x) {
  return toclf(x);
}

std::ostream& operator<<(std::ostream& os, const quadtype& x) {
  return os << tonumeric(x);
}

#endif

#ifdef QUAD_PRECISION

GiNaC::numeric tonumeric(const quadtype& x) {
  char buf[128];
  quadmath_snprintf(buf, sizeof buf, "%.35Qe", x);
  return numeric(buf);
}

double todouble(const quadtype& x) {
  return x;
}

inline quadtype toquad(const GiNaC::ex& x) {
  numeric n = torealnumeric(x);
  ostringstream ss;
  ss << n;
  quadtype q = strtoflt128(ss.str().c_str(), NULL);
  return q;
}

template<> quadtype qqvvamp<quadtype>::n(const GiNaC::ex& x) {
  return toquad(x);
}

std::ostream& operator<<(std::ostream& os, const cln::cl_F& x) {
  return os << tonumeric(x);
}

#endif

#ifdef DOUBLE_PRECISION

GiNaC::numeric tonumeric(const double& x) {
  return numeric(x);
}

double todouble(const double& x) {
  return x;
}

inline double todouble(const GiNaC::ex& x) {
  numeric n = torealnumeric(x);
  return n.to_double();
}

template<> double qqvvamp<double>::n(const GiNaC::ex& x) {
  return todouble(x);
}

#endif

namespace {
void qqvvamp_welcome() {
  static bool first = true;
  if (first) {
    first = false;
    cout << "\n\"The two-loop helicity amplitudes for qqb' -> V1V2 -> 4 leptons\"\n";
    cout << "T. Gehrmann, A. von Manteuffel, L. Tancredi\n";
    cout << "qqvvamp package, version 1.0\n" << endl;
  }
}
}

template<class T> qqvvamp<T>::qqvvamp(int Nf, IRscheme ir)
  : Nf(Nf), ir(ir), compute_gpls(true) {
  qqvvamp_welcome();
  r = new T[rsize];
  for (int j = 0 ; j < ntype; ++j) {
    for (int k = 0 ; k < nloop; ++k) {
      for (int l = 0 ; l < nreim ; ++l) {
	for (int i = 1 ; i <= nAcoeff; ++i)
	  A[i][j][k][l] = n(0);
	for (int i = 1 ; i <= nEcoeff; ++i)
	  E[i][j][k][l] = n(0);
      }
    }
  }
}

template<class T> qqvvamp<T>::~qqvvamp() {
  delete[] r;
}

namespace {
template<unsigned int n> inline numeric Pow( const numeric& x ) { return x*Pow<n-1>(x); }
template<> inline numeric Pow<0>( const numeric& x ) { return numeric(1); }
}

// for arbitrary precision type enforce calculation in specified precision
// even if input was constructed from low precision numbers
template<class T> numeric adjustdigits(const T& x) {
  return tonumeric(x);
}
template<> numeric adjustdigits(const cln::cl_F& x) {
  return numeric(cln::cl_float(x, cln::default_float_format));
}


// compute class [F] contributions
template<class T>
void compute_class_F(qqvvamp<T>& c, int Nf) {
  T (*n)(const ex&) = &qqvvamp<T>::n;

  ex Pi2 = Pi*Pi;
  ex Pi3 = Pi*Pi2;
  ex Pi4 = Pi2*Pi2;
  ex CF = numeric(4,3);
  ex N = numeric(3);
  
  ex f[nloop];
  
  f[0] = 1;

  f[1] = CF*( - numeric(4) + numeric(1,2)*Pi2 - numeric(3,2)*I*Pi );

  f[2] = CF*CF*( numeric(255,32) - numeric(23,12)*Pi2 + numeric(23,360)*Pi4
          - numeric(15,2)*zeta(3) + numeric(45,8)*I*Pi - numeric(1,4)*I*Pi3
          - numeric(6)*I*zeta(3)*Pi  )
       + N*CF*( - numeric(1535,96) + numeric(215,108)*Pi2 - numeric(1,120)*Pi4
          + numeric(59,9)*zeta(3) - numeric(2545,216)*I*Pi - numeric(11,72)*I*Pi3
          + numeric(13,2)*I*zeta(3)*Pi )
       + Nf*CF * ( numeric(127,48) - numeric(19,54)*Pi2 + numeric(4,9)*zeta(3)
          + numeric(209,108)*I*Pi + numeric(1,36)*I*Pi3 );

  T ff[nloop][nreim];
  for (int l = 0; l < nloop ; ++l) {
     ff[l][0] = n(real_part(f[l]));
     ff[l][1] = n(imag_part(f[l]));
  }

  int F = 3;
  for (int i = 1; i <= 6; ++i)
    for (int l = 0; l < nloop; ++l)
      for (int r = 0 ; r < nreim; ++r)
        c.A[i][F][l][r] = n(0);
  
  for (int l = 0; l < nloop; ++l) {
    for (int r = 0; r < nreim; ++r) {
      c.A[7][F][l][r]  = n(2)*ff[l][r];
      c.A[8][F][l][r]  = n(2)*ff[l][r];
      c.A[9][F][l][r]  = -ff[l][r];
      c.A[10][F][l][r] = -ff[l][r];
    }
  }
}


// convert from qt-subtraction to Catani's original subtraction scheme
template<class T>
void convert_A_to_Catani(qqvvamp<T>& c, int Nf) {
  T (*n)(const ex&) = &qqvvamp<T>::n;
  
  numeric CF = numeric(4,3);
  numeric CA = numeric(3);
  
  ex I1toCatani = CF*( numeric(-3)*zeta(2) + I*Pi*numeric(3,2));
  ex I2toCatani =
     Nf*CF*( numeric(41,81)+numeric(97,36)*zeta(2)+numeric(-17,36)*zeta(3)
           + I*Pi*(numeric(-65,108) + numeric(-1,6)*zeta(2)))
    +CA*CF*( numeric(-607,162)+numeric(-1181,72)*zeta(2)
            +numeric(187,72)*zeta(3)+numeric(105,16)*zeta(4)
      +I*Pi*(numeric(961,216)+numeric(11,12)*zeta(2)+numeric(-13,2)*zeta(3)))
    +CF*CF*( numeric(-27,4)*zeta(2)+numeric(45,4)*zeta(4)
      + I*Pi*(numeric(3,8)+numeric(-15,2)*zeta(2)+numeric(6)*zeta(3)));

  I1toCatani = I1toCatani.evalf();
  I2toCatani = I2toCatani.evalf();
  
  for (int i = 1; i <= nAcoeff; ++i) {
    for (int j = 0 ; j < ntype; ++j) {
      ex A0qt = tonumeric(c.A[i][j][0][0]) + I*tonumeric(c.A[i][j][0][1]);
      ex A1qt = tonumeric(c.A[i][j][1][0]) + I*tonumeric(c.A[i][j][1][1]);
      ex A2qt = tonumeric(c.A[i][j][2][0]) + I*tonumeric(c.A[i][j][2][1]);
      ex A0Catani = A0qt;
      ex A1Catani = A1qt + I1toCatani*A0qt;
      ex A2Catani = A2qt + I2toCatani*A0qt + I1toCatani*A1qt;
      c.A[i][j][0][0] = n(real_part(A0Catani));
      c.A[i][j][0][1] = n(imag_part(A0Catani));
      c.A[i][j][1][0] = n(real_part(A1Catani));
      c.A[i][j][1][1] = n(imag_part(A1Catani));
      c.A[i][j][2][0] = n(real_part(A2Catani));
      c.A[i][j][2][1] = n(imag_part(A2Catani));
    }
  }
}

// generate the 9 E coefficients from the 10 A coefficients
template<class T>
void compute_ecoeff(qqvvamp<T>& c, numeric s, numeric t, numeric u, numeric ma2, numeric mb2) {
  T (*n)(const ex&) = &qqvvamp<T>::n;
  for (int j = 0 ; j < ntype; ++j) {
    for (int l = 0 ; l < nloop; ++l) {
      for (int m = 0; m < nreim; ++m) {
	numeric A[nAcoeff + 1];
	for (int i = 1; i <= nAcoeff; ++i)
	  A[i] = tonumeric(c.A[i][j][l][m]); // for better readability
	c.E[1][j][l][m] = n( A[1] );
	c.E[2][j][l][m] = n( A[2] + 2*A[9]/s - 2*A[10]/s );
	c.E[3][j][l][m] = n( A[3] - 2*A[9]/s + 2*A[10]/s );
	c.E[4][j][l][m] = n( A[4] );
	c.E[5][j][l][m] = n( 2*A[9] + 2*A[10] );
	c.E[6][j][l][m] = n( 2*A[7] + 2*A[9]*(u-ma2)/s - 2*A[10]*(u-ma2)/s );
	c.E[7][j][l][m] = n( 2*A[8] - 2*A[9]*(t-ma2)/s + 2*A[10]*(t-ma2)/s );
	c.E[8][j][l][m] = n( 2*A[5] - 2*A[9]*(u-ma2-s)/s - 2*A[10]*(t-mb2)/s );
	c.E[9][j][l][m] = n( 2*A[6] - 2*A[9]*(u-mb2)/s - 2*A[10]*(t-ma2-s)/s );
      }
    }
  }
}

template<class T>
void qqvvamp<T>::compute(T s_n, T t_n, T ma2_n, T mb2_n) {

  // for convenience: prepare input with GiNaC::numeric
  numeric s = adjustdigits(s_n);
  numeric t = adjustdigits(t_n);
  numeric ma2 = adjustdigits(ma2_n);
  numeric mb2 = adjustdigits(mb2_n);
  numeric u = ma2 + mb2 - s - t;

  // check whether phase space point in physical region
  
  if (ma2 <= 0 || mb2 <= 0)
    throw runtime_error("illegal ma2 or mb2 value");
  
  numeric smin = Pow<2>(sqrt(ma2) + sqrt(mb2));
  if (s < smin)
    throw runtime_error("illegal s value");

  numeric root = sqrt(Pow<2>(s-ma2-mb2)-4*ma2*mb2); // kappa
  numeric t0 = (ma2 + mb2 - s - root)/2;
  numeric t1 = (ma2 + mb2 - s + root)/2;
  if (t < t0 || t > t1) {
    cerr << "t value should be within [" << t0 << ", " << t1 << "]" << endl;
    throw runtime_error("illegal t value");
  }

  // prepare input to generated code
  numeric x =  -t0 / ma2 ;
  numeric y =  t1*(u+t)/(ma2*mb2) - 1 ; 
  numeric z =  t1*t / (ma2*mb2) ;
 
  bool verbose = false;
  if (verbose) {
    cout << "s = " << s << "\t" << cln::float_precision(As(cln::cl_F)(s.to_cl_N())) << "\n";
    cout << "t = " << t << "\t" << cln::float_precision(As(cln::cl_F)(t.to_cl_N())) << "\n";
    cout << "u = " << u << "\t" << cln::float_precision(As(cln::cl_F)(u.to_cl_N())) << "\n";
    cout << "ma2 = " << ma2 << "\t" << cln::float_precision(As(cln::cl_F)(ma2.to_cl_N())) << "\n";
    cout << "mb2 = " << mb2 << "\t" << cln::float_precision(As(cln::cl_F)(mb2.to_cl_N())) << "\n";
    cout << "x = " << x << "\t" << cln::float_precision(As(cln::cl_F)(x.to_cl_N())) << "\n";
    cout << "y = " << y << "\t" << cln::float_precision(As(cln::cl_F)(y.to_cl_N())) << "\n";
    cout << "z = " << z << "\t" << cln::float_precision(As(cln::cl_F)(z.to_cl_N())) << endl;
  }
  
  p0001(n(ma2), n(t/ma2), n(u/ma2), n(mb2/ma2), n(x), n(y), n(z));
  
  compute_class_F(*this, Nf);
  
  switch (ir) {
    case IRschemeQt:
      break;
    case IRschemeCatani:
      convert_A_to_Catani(*this, Nf);
      break;
  }

  compute_ecoeff(*this, s, t, u, ma2, mb2);
}



#include "qqvvamp.xi"
